from dataclasses import dataclass

class Type:
    DISK = "disk"
    FTP = "ftp"
    S3 = "s3"

@dataclass
class Key:
    url: str
    username: str
    password: str

@dataclass
class Resource:
    type: Type
    key: Key
    path: str


from .entities import Type, Key

from .storage import Storage
from .ftp import FtpServer
from .s3 import S3

def new_storage_by_type(storage_type: Type, key: Key) -> Storage:
	# if storage_type == Type.DISK:
	# 	return NewDisk()

    print(key)
	
    if storage_type == Type.FTP:
        return FtpServer(key.url, key.username, key.password)
    if storage_type == Type.S3:
        return S3(key.url, key.username, key.password)
        
    return None


"""Абстрактный класс для работы с хранилищами"""
from abc import ABC, abstractmethod
from enum import Enum
from dataclasses import dataclass
from datetime import datetime

class EntryType(Enum):
    """EntryType описывает тип файла"""
    FILE = 1
    FOLDER = 2
    LINK = 3


@dataclass
class FileData:
    """FileData описывает файл ftp-сервера"""
    name: str
    size: int = 0
    etype: EntryType = EntryType.FILE
    time: datetime = datetime.now()

    def __init__(
        self, name: str, size: int = 0,
        etype: EntryType = EntryType.FILE,
        time: datetime = datetime.now(),
    ) -> None:
        self.name = name
        self.size = size
        self.etype = etype
        self.time = time


class Storage(ABC):
    """Storage базовый класс для работы с хранилищами"""
    @abstractmethod
    def files(self, path: str) -> bool:
        """upload загружает файл source в Storage"""

    @abstractmethod
    def upload(self, source: str, destination: str) -> bool:
        """upload загружает файл source в Storage"""

    @abstractmethod
    def download(self, source: str, destination: str) -> bool:
        """download скачивает файл source из Storage"""



"""Класс для композиции Storage"""
from .storage import Storage, FileData


class StorageComposite(Storage):
    """StorageComposite позволяет создавать композицию Storage"""
    def __init__(self) -> None:
        self._storages: [Storage] = []

    def add(self, storage: Storage) -> None:
        """add добавляет новый Storage"""
        self._storages.append(storage)

    def remove(self, storage: Storage) -> None:
        """remove удаляет существующий Storage"""
        self._storages.remove(storage)

    def files(self, path: str) -> [FileData]:
        """files возвращает список файлов из всех Storage"""
        files = []
        for storage in self._storages:
            files.append(storage.files(path))
        return files

    def upload(self, source: str, destination: str) -> bool:
        """
        upload загружает файл source во все Storage
        возвращает True если запись была произведена во все Storage
        """
        uploaded = True
        for storage in self._storages:
            if not storage.upload(source, destination):
                uploaded = False
        return uploaded

    def download(self, source: str, destination: str) -> bool:
        """
        download скачивает файл source из всех Storage
        возвращает True если файл был скачан хотя бы из одного Storage
        """
        downloaded = False
        for storage in self._storages:
            if storage.download(source, destination):
                downloaded = True
        return downloaded



"""Пакет для работы с ftp-сервером"""
from ftplib import FTP, Error
from datetime import datetime

from .storage import Storage, FileData, EntryType


class FtpServer(Storage):
    """FtpServer класс взаимодействия с ftp-сервером"""
    def __init__(self, host: str, user: str, password: str) -> None:
        self._ftp = None
        self._host = host
        self._user = user
        self._password = password

    def _connect(self) -> bool:
        """_connect подключает к ftp-серверу"""
        self._ftp = FTP(self._host)
        try:
            if self._ftp.login(self._user, self._password) != "230 Login successful.":
                return False
            return True
        except Error:
            return False

    def _disconnect(self) -> None:
        """_disconnect отключает от ftp-сервера"""
        if self._ftp is not None:
            try:
                self._ftp.quit()
            except Error:
                self._ftp.close()
        self._ftp = None

    @staticmethod
    def _decode_file_data(data: str) -> dict:
        """_decode_file_data декодирует данные о файлах на ftp-сервере"""
        fields = data.split()
        return {
            "type": EntryType.FOLDER if fields[0] == "d" else EntryType.FILE,
            "size": int(fields[4]),
            "name": fields[-1],
            "time": datetime(   # время изменения
                year=datetime.now().year,
                month=datetime.strptime(fields[5], "%b").month,
                day=int(fields[6]),
                hour=int(fields[7].split(":")[0]),
                minute=int(fields[7].split(":")[-1]),
            ),
        }

    def files(self, path: str) -> []:
        """files получает информацию о файлах в директории path на ftp-сервера"""
        files = []
        if not self._connect():
            return files
        try:
            notes = []
            self._ftp.cwd(path)
            self._ftp.retrlines("LIST", notes.append)
            for note in notes:
                data = self._decode_file_data(note)
                files.append(FileData(data["name"], data["size"], data["type"], data["time"]))
            self._disconnect()
            return files
        except (FileNotFoundError, Error):
            self._disconnect()
            return files

    def upload(self, source: str, destination: str) -> bool:
        """upload загружает файл source на ftp-сервер"""
        if not self._connect():
            return False
        filename = basename(source)
        try:
            if filename == "":
                for filename in listdir(source):
                    with open(join(source, filename), "rb") as file:
                        self._ftp.storbinary("STOR " + join(destination, filename), file, 1024)
            else:
                if basename(destination) == "":
                    destination = join(destination, filename)
                with open(source, "rb") as file:
                    self._ftp.storbinary("STOR " + destination, file, 1024)
            self._disconnect()
            return True
        except (FileNotFoundError, Error):
            self._disconnect()
            return False

    def download(self, source: str, destination: str) -> bool:
        """download скачивает файл source с ftp-сервера"""
        if not self._connect():
            return False
        filename = basename(source)
        try:
            if filename == "":
                for filedata in self.files(source):
                    with open(join(destination, filedata.name), "wb") as file:
                        self._ftp.retrbinary("RETR " + join(source, filedata.name), file.write, 1024)
            else:
                if basename(destination) == "":
                    destination = join(destination, filename)
                with open(destination, "wb") as file:
                    self._ftp.retrbinary("RETR " + source, file.write, 1024)
            self._disconnect()
            return True
        except (FileNotFoundError, Error):
            self._disconnect()
            return False
